/* Name: main.c
 * Project: HID-Test
 * Author: Christian Starkjohann
 * Creation Date: 2006-02-02
 * Tabsize: 4
 * Copyright: (c) 2006 by OBJECTIVE DEVELOPMENT Software GmbH
 * License: GNU GPL v2 (see License.txt) or proprietary (CommercialLicense.txt)
 * This Revision: $Id: main.c 299 2007-03-29 17:07:19Z cs $
 */

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>
#include <util/delay.h>

#include <stdio.h>
#include <string.h>

#include "usbdrv.h"
#include "oddebug.h"

#include "FileSys/ff.h"
#include "FileSys/mmc.h"

#define SWVERSIONMAJOR 0
#define SWVERSIONMINOR 1
#define DEVID 2011


#define MAXMACRO_LEN	150
#define ACTIVATE_USB	1
#define ENABLE_FS		1
#define NO_CRLF			1				// CRLF is not reported to PC , string "\n" is used in macros to send a CRLF

// test environment
#define TEST_CODE		0
#define TEST_STRINGS	0

#define ERROR_PATTERN1	0b1010101000000000
#define ERROR_PATTERN2	0b0101010100000000
#define ERROR_BLINK		10000		// in 1/10 ms

// error codes

#define 	ERR_NOCARD		0x01
#define 	ERR_NOMACROTXT	0x02
#define 	ERR_INVMACRO	0x04
#define 	ERR_FSERROR		0x08


/* ----------------------- variable definition ----------------------------- */

// filesystem 
FATFS MMCCard;
FIL f;

// usb	HID 
static uchar    reportBuffer[2];    /* buffer for HID reports */
static uchar    idleRate;           /* in 4 ms units */

// keys and leds
static WORD s_wActiveLed=0;
static BYTE wCurrRow=16;		// 
static WORD wCurrLed=s_wActiveLed;

// string buffer
char s_szTempBuffer[MAXMACRO_LEN];


/* ----------------------- bootloader information -------------------------- */

typedef struct
{
	unsigned long dev_id;
	unsigned short app_version;
	unsigned short crc;
} bootldrinfo_t;

const bootldrinfo_t bootlodrinfo __attribute__ ((section (".bootldrinfo"))) = {DEVID, SWVERSIONMAJOR << 8 | SWVERSIONMINOR, 0x0000};


/* ----------------------- hardware I/O abstraction ------------------------ */

/* pin assignments:
PB0	Key 1
PB1	Key 2
PB2	Key 3
PB3	Key 4
PB4	Key 5
PB5 Key 6

PC0	Key 7
PC1	Key 8
PC2	Key 9
PC3	Key 10
PC4	Key 11
PC5	Key 12

PD0	USB-
PD1	debug tx
PD2	USB+ (int0)
PD3	Key 13
PD4	Key 14
PD5	Key 15
PD6	Key 16
PD7	Key 17
*/

void ShowVersion();
void ShowError(char nErr,char nSubErr=0);
void SendString(const char *pStr,char bShowCLRF=0);

static void hardwareInit(void)
{
uchar	i, j;

    PORTD = 0xfa;   /* 1111 1010 bin: activate pull-ups except on USB lines */
    DDRD = 0x07;    /* 0000 0111 bin: all pins input except USB (-> USB reset) */
	j = 0;
	while(--j){     /* USB Reset by device only required on Watchdog Reset */
		i = 0;
		while(--i); /* delay >10ms for USB reset */
	}
    DDRD = 0x02;    /* 0000 0010 bin: remove USB reset condition */
    /* configure timer 0 for a rate of 12M/(1024 * 256) = 45.78 Hz (~22ms) */
    TCCR0A = 5;      /* timer 0 prescaler: 1024 */
}

/* ------------------------------------------------------------------------- */
// pc4,pc5,pb0,pb1  led cathods
// pc0..pc3 		led anods


BYTE ReadKey()
{
	BYTE nRow;
	BYTE nPin;

	// pc0..pc3 here as outputs: low
	BYTE nPinState= (PINC & 0x30) | (PINB & 0x03) ;
	switch (nPinState) {
		case 0b00110010: nRow=0; break;
		case 0b00110001: nRow=1; break;
		case 0b00100011: nRow=2; break;
		case 0b00010011: nRow=3; break;
			// xx..xx..		<- significant pins
		default:
			return 0;
	}

	BYTE nCol=1;
	BYTE nColMsk=0xe;
	for (;;) {
		DDRC = 0xf & nColMsk;		// disable output for all cols, but one
		_delay_us(20);
		nPin = (PINC & 0x30) | (PINB & 0x03) ;	// read new pinstate
		if (nPin != nPinState)					// changed -> found affected column
			break;
		nColMsk = nColMsk<<1;
		nCol++;
	} 
	return (nRow<<2)+nCol;
}

BYTE KeyLEDCyclic()
{
	static BYTE nKey1=0xff;
	static BYTE nLastKey=0xff;
	BYTE bKeyFound=0;
	BYTE nKey;
	PORTC = (wCurrLed & 0x0f) | ((~wCurrRow & 0xc)<<2) ;	
	PORTB = (~wCurrRow) & 0x3 ;	

	if (16 == wCurrRow) {
		// prepare keyboard read
		DDRC=0xf;
		DDRB=0;
		PORTB = 3;	
		PORTC = 0x30;		// pullup active
		_delay_us(20);
			
		nKey=ReadKey();
		if (nKey) {
        	if(nKey1 == nKey && nLastKey != nKey){
	            nLastKey = nKey;
				bKeyFound=1;	            
        	}
			nKey1=nKey;
		} 
		else {
			if( 0xff == nKey1 )
				nLastKey=0xff;
			nKey1=0xff;
		}
		
		// prepare next LED output
		DDRC=0x3f;
		DDRB=0x3;
		wCurrRow=1;
		wCurrLed=s_wActiveLed;
	}
	else {
		wCurrRow <<= 1;
		wCurrLed >>= 4;
	}
	return bKeyFound? nLastKey : 0;
}

/* ------------------------------------------------------------------------- */
/* ----------------------------- USB interface ----------------------------- */
/* ------------------------------------------------------------------------- */

PROGMEM char usbHidReportDescriptor[35] = { /* USB report descriptor */
    0x05, 0x01,                    // USAGE_PAGE (Generic Desktop)
    0x09, 0x06,                    // USAGE (Keyboard)
    0xa1, 0x01,                    // COLLECTION (Application)
    0x05, 0x07,                    //   USAGE_PAGE (Keyboard)
    0x19, 0xe0,                    //   USAGE_MINIMUM (Keyboard LeftControl)
    0x29, 0xe7,                    //   USAGE_MAXIMUM (Keyboard Right GUI)
    0x15, 0x00,                    //   LOGICAL_MINIMUM (0)
    0x25, 0x01,                    //   LOGICAL_MAXIMUM (1)
    0x75, 0x01,                    //   REPORT_SIZE (1)
    0x95, 0x08,                    //   REPORT_COUNT (8)
    0x81, 0x02,                    //   INPUT (Data,Var,Abs)
    0x95, 0x01,                    //   REPORT_COUNT (1)
    0x75, 0x08,                    //   REPORT_SIZE (8)
    0x25, 0x65,                    //   LOGICAL_MAXIMUM (101)
    0x19, 0x00,                    //   USAGE_MINIMUM (Reserved (no event indicated))
    0x29, 0x65,                    //   USAGE_MAXIMUM (Keyboard Application)
    0x81, 0x00,                    //   INPUT (Data,Ary,Abs)
    0xc0                           // END_COLLECTION
};
/* We use a simplifed keyboard report descriptor which does not support the
 * boot protocol. We don't allow setting status LEDs and we only allow one
 * simultaneous key press (except modifiers). We can therefore use short
 * 2 byte input reports.
 * The report descriptor has been created with usb.org's "HID Descriptor Tool"
 * which can be downloaded from http://www.usb.org/developers/hidpage/.
 * Redundant entries (such as LOGICAL_MINIMUM and USAGE_PAGE) have been omitted
 * for the second INPUT item.
 */

/* Keyboard usage values, see usb.org's HID-usage-tables document, chapter
 * 10 Keyboard/Keypad Page for more codes.
 */
#define MOD_NULL   			(0)
#define MOD_CONTROL_LEFT    (1<<0)
#define MOD_SHIFT_LEFT      (1<<1)
#define MOD_ALT_LEFT        (1<<2)
#define MOD_GUI_LEFT        (1<<3)
#define MOD_CONTROL_RIGHT   (1<<4)
#define MOD_SHIFT_RIGHT     (1<<5)
#define MOD_ALT_RIGHT       (1<<6)
#define MOD_GUI_RIGHT       (1<<7)

#define MOD_ALTGR	(MOD_CONTROL_LEFT | MOD_ALT_LEFT)

/*
	0x00	Reserved (no event indicated)
	0x01	Keyboard ErrorRollOver
	0x02	Keyboard POSTFail
	0x03	Keyboard ErrorUndefined
	0x04	Keyboard a and A
	0x05	Keyboard b and B
	0x06	Keyboard c and C
	0x07	Keyboard d and D
	0x08	Keyboard e and E
	0x09	Keyboard f and F
	0x0A	Keyboard g and G
	0x0B	Keyboard h and H
	0x0C	Keyboard i and I
	0x0D	Keyboard j and J
	0x0E	Keyboard k and K
	0x0F	Keyboard l and L
	0x10	Keyboard m and M
	0x11	Keyboard n and N
	0x12	Keyboard o and O
	0x13	Keyboard p and P
	0x14	Keyboard q and Q
	0x15	Keyboard r and R
	0x16	Keyboard s and S
	0x17	Keyboard t and T
	0x18	Keyboard u and U
	0x19	Keyboard v and V
	0x1A	Keyboard w and W
	0x1B	Keyboard x and X
	0x1C	Keyboard y and Y
	0x1D	Keyboard z and Z
	0x1E	Keyboard 1 and !
	0x1F	Keyboard 2 and @
	0x20	Keyboard 3 and #
	0x21	Keyboard 4 and $
	0x22	Keyboard 5 and %
	0x23	Keyboard 6 and ^
	0x24	Keyboard 7 and &
	0x25	Keyboard 8 and *
	0x26	Keyboard 9 and (
	0x27	Keyboard 0 and )
	0x28	Keyboard Return (ENTER)
	0x29	Keyboard ESCAPE
	0x2A	Keyboard DELETE (Backspace)
	0x2B	Keyboard Tab
	0x2C	Keyboard Spacebar
	0x2D	Keyboard - and (underscore)
	0x2E	Keyboard = and +
	0x2F	Keyboard [ and {
	0x30	Keyboard ] and }
	0x31	Keyboard \ and |
	0x32	Keyboard Non-US # and ~
	0x33	Keyboard ; and :
	0x34	Keyboard ' and "
	0x35	Keyboard Grave Accent and Tilde
	0x36	Keyboard, and <
	0x37	Keyboard . and >
	0x38	Keyboard / and ?
	0x39	Keyboard Caps Lock
	0x3A	Keyboard F1
	0x3B	Keyboard F2
	0x3C	Keyboard F3
	0x3D	Keyboard F4
	0x3E	Keyboard F5
	0x3F	Keyboard F6
	0x40	Keyboard F7
	0x41	Keyboard F8
	0x42	Keyboard F9
	0x43	Keyboard F10
	0x44	Keyboard F11
	0x45	Keyboard F12
	0x46	Keyboard PrintScreen
	0x47	Keyboard Scroll Lock
	0x48	Keyboard Pause
	0x49	Keyboard Insert
	0x4A	Keyboard Home
	0x4B	Keyboard PageUp
	0x4C	Keyboard Delete Forward
	0x4D	Keyboard End
	0x4E	Keyboard PageDown
	0x4F	Keyboard RightArrow
	0x50	Keyboard LeftArrow
	0x51	Keyboard DownArrow
	0x52	Keyboard UpArrow
	0x53	Keypad Num Lock and Clear
	0x54	Keypad /
	0x55	Keypad *
	0x56	Keypad -
	0x57	Keypad +
	0x58	Keypad ENTER
	0x59	Keypad 1 and End
	0x5A	Keypad 2 and Down Arrow
	0x5B	Keypad 3 and PageDn
	0x5C	Keypad 4 and Left Arrow
	0x5D	Keypad 5
	0x5E	Keypad 6 and Right Arrow
	0x5F	Keypad 7 and Home
	0x60	Keypad 8 and Up Arrow
	0x61	Keypad 9 and PageUp
	0x62	Keypad 0 and Insert
	0x63	Keypad . and Delete
	0x64	Keyboard Non-US \ and |
	0x65	Keyboard Application
	0x66	Keyboard Power
	0x67	Keypad =
	0x68	Keyboard F13
	0x69	Keyboard F14
	0x6A	Keyboard F15
	0x6B	Keyboard F16
	0x6C	Keyboard F17
	0x6D	Keyboard F18
	0x6E	Keyboard F19
	0x6F	Keyboard F20
	0x70	Keyboard F21
	0x71	Keyboard F22
	0x72	Keyboard F23
	0x73	Keyboard F24
	0x74	Keyboard Execute
	0x75	Keyboard Help
	0x76	Keyboard Menu
	0x77	Keyboard Select
	0x78	Keyboard Stop
	0x79	Keyboard Again
	0x7A	Keyboard Undo
	0x7B	Keyboard Cut
	0x7C	Keyboard Copy
	0x7D	Keyboard Paste
	0x7E	Keyboard Find
	0x7F	Keyboard Mute
	0x80	Keyboard Volume Up
	0x81	Keyboard Volume Down
	0x82	Keyboard Locking Caps Lock
	0x83	Keyboard Locking Num Lock
	0x84	Keyboard Locking Scroll Lock
	0x85	Keypad Comma
	0x86	Keypad Equal Sign
	0x87	Keyboard International1
	0x88	Keyboard International2
	0x89	Keyboard International3
	0x8A	Keyboard International4
	0x8B	Keyboard International5
	0x8C	Keyboard International6
	0x8D	Keyboard International7
	0x8E	Keyboard International8
	0x8F	Keyboard International9
	0x90	Keyboard LANG1
	0x91	Keyboard LANG2
	0x92	Keyboard LANG3
	0x93	Keyboard LANG4
	0x94	Keyboard LANG5
	0x95	Keyboard LANG6
	0x96	Keyboard LANG7
	0x97	Keyboard LANG8
	0x98	Keyboard LANG9
	0x99	Keyboard Alternate Erase
	0x9A	Keyboard SysReq/Attention
	0x9B	Keyboard Cancel
	0x9C	Keyboard Clear
	0x9D	Keyboard Prior
	0x9E	Keyboard Return
	0x9F	Keyboard Separator
	0xA0	Keyboard Out
	0xA1	Keyboard Oper
	0xA2	Keyboard Clear/Again
	0xA3	Keyboard CrSel/Props
	0xA4	Keyboard ExSel
	0xE0	Keyboard LeftControl
	0xE1	Keyboard LeftShift
	0xE2	Keyboard LeftAlt
	0xE3	Keyboard Left GUI
	0xE4	Keyboard RightControl
	0xE5	Keyboard RightShift
	0xE6	Keyboard RightAlt
	0xE7	Keyboard Right GUI 
*/

static const uchar  keyReport[][2] PROGMEM = {
/* NUL*/    {MOD_NULL	   , 0},		// 0
/* SOH*/    {MOD_NULL	   , 0},		// 1
/* STX*/    {MOD_NULL	   , 0},		// 2
/* ETX*/    {MOD_NULL	   , 0},		// 3
/* EOT*/    {MOD_NULL	   , 0},		// 4
/* ENQ*/    {MOD_NULL	   , 0},		// 5
/* ACK*/    {MOD_NULL	   , 0},		// 6
/* BEL*/    {MOD_NULL	   , 0},		// 7
/* BS */    {MOD_NULL	   , 0x2a},		// 8
/* TAB*/    {MOD_NULL	   , 0x2B},		// 9
/* LF */    {MOD_NULL	   , 0x28},		// a
/* VT */    {MOD_NULL	   , 0},		// b 
/* FF */    {MOD_NULL	   , 0},		// c 
/* CR */    {MOD_NULL	   , 0x28},		// d 
/* SO */    {MOD_NULL	   , 0},		// e 
/* SI */    {MOD_NULL	   , 0},		// f 
/* DLE*/    {MOD_NULL	   , 0},		// 10
/* DC1*/    {MOD_NULL	   , 0},		// 11
/* DC2*/    {MOD_NULL	   , 0},		// 12
/* DC3*/    {MOD_NULL	   , 0},		// 13
/* DC4*/    {MOD_NULL	   , 0},		// 14
/* NAK*/    {MOD_NULL	   , 0},		// 15
/* SYN*/    {MOD_NULL	   , 0},		// 16
/* ETB*/    {MOD_NULL	   , 0},		// 17
/* CAN*/    {MOD_NULL	   , 0},		// 18
/* EM */    {MOD_NULL	   , 0},		// 19
/* SUB*/    {MOD_NULL	   , 0},		// 1a
/* ESC*/    {MOD_NULL	   , 0x29},		// 1b	
/* FS */    {MOD_NULL	   , 0},		// 1c
/* GS */    {MOD_NULL	   , 0},		// 1d
/* RS */    {MOD_NULL	   , 0},		// 1e
/* US */    {MOD_NULL	   , 0},		// 1f

/*    */    {MOD_NULL	   , 0x2c},		// 20
/*  ! */    {MOD_SHIFT_LEFT, 30},		// 21
/*  " */    {MOD_SHIFT_LEFT, 31},		// 22
/*  # */    {MOD_NULL	   , 50},		// 23
/*  $ */    {MOD_SHIFT_LEFT, 33},		// 24
/*  % */    {MOD_SHIFT_LEFT, 34},		// 25
/*  & */    {MOD_SHIFT_LEFT, 35},		// 26
/*  ' */    {MOD_SHIFT_LEFT, 50},		// 27
/*  ( */    {MOD_SHIFT_LEFT, 37},		// 28
/*  ) */    {MOD_SHIFT_LEFT, 38},		// 29

/*  * */    {MOD_SHIFT_LEFT, 48},		// 2a
/*  + */    {MOD_NULL	   , 48},		// 2b
/*  , */    {MOD_NULL	   , 54},		// 2c
/*  - */    {MOD_NULL      , 56},		// 2d
/*  . */    {MOD_NULL	   , 55},		// 2e
/*  / */    {MOD_SHIFT_LEFT, 36},		// 2f

/*  0 */    {MOD_NULL	   , 39},		// 30
/*  1 */    {MOD_NULL	   , 30},		// 31
/*  2 */    {MOD_NULL	   , 31},		// 32
/*  3 */    {MOD_NULL	   , 32},		// 33
/*  4 */    {MOD_NULL	   , 33},		// 34
/*  5 */    {MOD_NULL	   , 34},		// 35
/*  6 */    {MOD_NULL	   , 35},		// 36
/*  7 */    {MOD_NULL	   , 36},		// 37
/*  8 */    {MOD_NULL	   , 37},		// 38
/*  9 */    {MOD_NULL	   , 38},		// 39


/*  : */    {MOD_SHIFT_LEFT ,55},		// 3a
/*  ; */    {MOD_SHIFT_LEFT, 54},		// 3b
/*  < */    {MOD_NULL	   , 100},		// 3c
/*  = */    {MOD_SHIFT_LEFT, 39},		// 3d
/*  > */    {MOD_SHIFT_LEFT, 100},		// 3e
/*  ? */    {MOD_SHIFT_LEFT, 45},		// 3f

/*  @ */    {MOD_ALTGR	   , 20},		// 40
/*  A */    {MOD_SHIFT_LEFT, 4},		// 41
/*  B */    {MOD_SHIFT_LEFT, 5},		// 42
/*  C */    {MOD_SHIFT_LEFT, 6},		// 43
/*  D */    {MOD_SHIFT_LEFT, 7},		// 44
/*  E */    {MOD_SHIFT_LEFT, 8},		// 45
/*  F */    {MOD_SHIFT_LEFT, 9},		// 46
/*  G */    {MOD_SHIFT_LEFT, 10},		// 47
/*  H */    {MOD_SHIFT_LEFT, 11},		// 48
/*  I */    {MOD_SHIFT_LEFT, 12},		// 49
/*  J */    {MOD_SHIFT_LEFT, 13},		// 4a
/*  K */    {MOD_SHIFT_LEFT, 14},		// 4b
/*  L */    {MOD_SHIFT_LEFT, 15},		// 4c
/*  M */    {MOD_SHIFT_LEFT, 16},		// 4d
/*  N */    {MOD_SHIFT_LEFT, 17},		// 4e
/*  O */    {MOD_SHIFT_LEFT, 18},		// 4f
/*  P */    {MOD_SHIFT_LEFT, 19},		// 50
/*  Q */    {MOD_SHIFT_LEFT, 20},		// 51
/*  R */    {MOD_SHIFT_LEFT, 21},		// 52
/*  S */    {MOD_SHIFT_LEFT, 22},		// 53
/*  T */    {MOD_SHIFT_LEFT, 23},		// 54
/*  U */    {MOD_SHIFT_LEFT, 24},		// 55
/*  V */    {MOD_SHIFT_LEFT, 25},		// 56
/*  W */    {MOD_SHIFT_LEFT, 26},		// 57
/*  X */    {MOD_SHIFT_LEFT, 27},		// 58
/*  Y */    {MOD_SHIFT_LEFT, 29},		// 59
/*  Z */    {MOD_SHIFT_LEFT, 28},		// 5a

/*  [ */    {MOD_ALTGR	   , 37},		// 5b
/*  \ */    {MOD_ALTGR	   , 45},		// 5c
/*  ] */    {MOD_ALTGR	   , 38},		// 5d
/*  ^ */    {MOD_NULL	   , 53},		// 5e
/*  _ */    {MOD_SHIFT_LEFT, 56},		// 5f

/*  ` */    {MOD_SHIFT_LEFT, 46},		// 60
/*  a */    {MOD_NULL	   , 4},		// 61
/*  b */    {MOD_NULL	   , 5},		// 62
/*  c */    {MOD_NULL	   , 6},		// 63
/*  d */    {MOD_NULL	   , 7},		// 64
/*  e */    {MOD_NULL	   , 8},		// 65
/*  f */    {MOD_NULL	   , 9},		// 66
/*  g */    {MOD_NULL	   , 10},		// 67
/*  h */    {MOD_NULL	   , 11},		// 68
/*  i */    {MOD_NULL	   , 12},		// 69
/*  j */    {MOD_NULL	   , 13},		// 6a
/*  k */    {MOD_NULL	   , 14},		// 6b
/*  l */    {MOD_NULL	   , 15},		// 6c
/*  m */    {MOD_NULL	   , 16},		// 6d
/*  n */    {MOD_NULL	   , 17},		// 6e
/*  o */    {MOD_NULL	   , 18},		// 6f
/*  p */    {MOD_NULL	   , 19},		// 70
/*  q */    {MOD_NULL	   , 20},		// 71
/*  r */    {MOD_NULL	   , 21},		// 72
/*  s */    {MOD_NULL	   , 22},		// 73
/*  t */    {MOD_NULL	   , 23},		// 74
/*  u */    {MOD_NULL	   , 24},		// 75
/*  v */    {MOD_NULL	   , 25},		// 76
/*  w */    {MOD_NULL	   , 26},		// 77
/*  x */    {MOD_NULL	   , 27},		// 78
/*  y */    {MOD_NULL	   , 29},		// 79
/*  z */    {MOD_NULL	   , 28},		// 7a

/*  { */    {MOD_ALTGR	   , 36},		// 7b
/*  | */    {MOD_ALTGR	   , 100},		// 7c
/*  } */    {MOD_ALTGR	   , 39},		// 7d
/*  ~ */    {MOD_ALTGR	   , 48},		// 7e

/* NUL*/    {MOD_NULL	   , 0},		// 7f
		
/*   */    {MOD_NULL	   , 45},		// 80
/*   */    {MOD_NULL	   , 47},		// 81
/*   */    {MOD_SHIFT_LEFT, 47},		// 82
/*   */    {MOD_NULL	   , 51},		// 83
/*   */    {MOD_SHIFT_LEFT, 51},		// 84
/*   */    {MOD_NULL	   , 52},		// 85
/*   */    {MOD_SHIFT_LEFT, 52},		// 86

/*   */    {MOD_SHIFT_LEFT, 32},		// 87
/*   */    {MOD_ALTGR	   , 31},		// 88
/*   */    {MOD_ALTGR     , 32},		// 89
/*   */    {MOD_ALTGR	   , 16},		// 8a
/*	euro */	{MOD_ALTGR	   , 8},		// 8b
	
/*  ^ */    {MOD_NULL	   , 53},		// 8c
/*   */    {MOD_SHIFT_LEFT, 53},		// 8d

};

static const uchar germanChars[] PROGMEM = {		// ħ^		 = euro
    0xDF, 0xFC, 0xDC, 0xF6, 0xD6, 0xE4, 0xC4, 0xA7, 0xB2, 0xB3, 0xB5, 0x80, 0x5E, 0xB0, 0x00
} ;

static void buildReport(uchar key)
{
	const prog_char *pC;
	const prog_char *pGC=(const prog_char*)germanChars;
	if (0x80 <= key) {		// special handling for german characters
		pC=strchr_P(pGC,key);
		if (NULL==pC) {
			*(int *)reportBuffer = 0;
			return;
		}
		key=pC-pGC + 0x80;
	}

	/* This (not so elegant) cast saves us 10 bytes of program memory */
    *(int *)reportBuffer = pgm_read_word(keyReport[key]);
}

static void buildNullReport()
{
/* This (not so elegant) cast saves us 10 bytes of program memory */
    *(int *)reportBuffer = 0;
}

uchar	usbFunctionSetup(uchar data[8])
{
	usbRequest_t    *rq = (usbRequest_t    *)data;

    usbMsgPtr = reportBuffer;
    if((rq->bmRequestType & USBRQ_TYPE_MASK) == USBRQ_TYPE_CLASS){    // class request type
        if(rq->bRequest == USBRQ_HID_GET_REPORT){  // wValue: ReportType (highbyte), ReportID (lowbyte)
            // we only have one report type, so don't look at wValue
            buildReport(0/*keyPressed()*/);
            return sizeof(reportBuffer);
        }else if(rq->bRequest == USBRQ_HID_GET_IDLE){
            usbMsgPtr = &idleRate;
            return 1;
        }else if(rq->bRequest == USBRQ_HID_SET_IDLE){
            idleRate = rq->wValue.bytes[1];
        }
    }else{
        /// no vendor specific requests implemented
    }
	return 0;
}

// find sring for stringID and copy it to s_szTempBuffer
uchar FindString(const char *psczStringID)
{
#if ENABLE_FS
	//SPI_USEDBY(eUndefined);
	s_SPIUsedBy=eUndefined;
	if (FR_OK==f_open(&f,"macros.txt",FA_READ)) {
		while (!f_eof(&f)) {
			if (0==f_gets(s_szTempBuffer,sizeof(s_szTempBuffer),&f))
				break;
			if (0==strncmp(s_szTempBuffer,psczStringID,2)) {
				unsigned int nLength=strlen(s_szTempBuffer+3);
				strncpy(s_szTempBuffer,s_szTempBuffer+3,nLength);
				s_szTempBuffer[nLength]=0;
				f_close(&f);
				return 1;
			}

		}
		f_close(&f);
	}
#endif
	return 0;
}

void SendChar(char nChar)
{
#if ACTIVATE_USB
	unsigned int nTimeout=20000;
    buildReport(nChar);
    usbSetInterrupt(reportBuffer, sizeof(reportBuffer));
	while (!usbInterruptIsReady() && nTimeout) {
		wdt_reset();
		usbPoll();
		nTimeout--;
	}
	buildNullReport();
	usbSetInterrupt(reportBuffer, sizeof(reportBuffer));
	while (!usbInterruptIsReady() && nTimeout) {
		wdt_reset();
		usbPoll();
		nTimeout--;
	}
#endif
}



#if TEST_CODE && ACTIVATE_USB

void TestSendChar(uchar nChar,uchar nShift=0)
{
	while (!usbInterruptIsReady())
		usbPoll();
	reportBuffer[0]=nShift;
	reportBuffer[1]=nChar;
	usbSetInterrupt(reportBuffer, sizeof(reportBuffer));
}

void TestKeyKode(int n)
{
	TestSendChar(0x28,0);		// enter
	sprintf(s_szTempBuffer,"%2i !",n);
	SendString(s_szTempBuffer);

	TestSendChar(n,0);
	TestSendChar(0,0);

	TestSendChar(n,MOD_SHIFT_LEFT);
	TestSendChar(0,0);

	TestSendChar(n,MOD_ALTGR);
	TestSendChar(0,0);
}
#endif

void wdt_init(void) __attribute__((naked)) __attribute__((section(".init3")));

// disable watchdog at startup
void wdt_init(void)
{
    MCUSR = 0;
 //   wdt_disable();
    return;
}

/* ------------------------------------------------------------------------- */

void ShowVersion()
{
	sprintf_P(s_szTempBuffer,PSTR("MacroPad 2010\n"));
	SendString(s_szTempBuffer,1);
	sprintf_P(s_szTempBuffer,PSTR("Version %i.%i\n"),SWVERSIONMAJOR,SWVERSIONMINOR);
	SendString(s_szTempBuffer,1);
	sprintf_P(s_szTempBuffer,PSTR("Build %s\n"),__DATE__);
	SendString(s_szTempBuffer,1);
	sprintf_P(s_szTempBuffer,PSTR("DeviceID: %i\n"),DEVID);
	SendString(s_szTempBuffer,1);
	sprintf_P(s_szTempBuffer,PSTR("c Th. Horn \n"));
	SendString(s_szTempBuffer,1);
	s_szTempBuffer[0]=0;
}

/* ------------------------------------------------------------------------- */

void ShowError(char nErr,char nSubErr)
{
	int key;
	int nCycles=0;

	s_wActiveLed = ERROR_PATTERN1 | nErr;
	for(;;){
		nCycles++;
		wdt_reset();
		usbPoll();
		key = KeyLEDCyclic();
		if (nCycles > ERROR_BLINK/2)
			s_wActiveLed = ERROR_PATTERN2 | nErr;
		if (nCycles > ERROR_BLINK) {
			s_wActiveLed = ERROR_PATTERN1 | nErr;
			nCycles=0;
		}
		switch (key) {
			case 1:
				sprintf_P(s_szTempBuffer,PSTR("\nError %i.%i\n"),nErr,nSubErr);
				SendString(s_szTempBuffer,1);
				break;
			case 2:
				ShowVersion();
				break;
		}
		_delay_us(100);

		if ( (ERR_NOCARD!=nErr) && (0!= (PIND & (1<<3)) )) {		// sd_card removed
			ShowError(ERR_NOCARD);
			return;
		}
		if ( (ERR_NOCARD==nErr) && (0== (PIND & (1<<3)) )) {		// check for re-inserted card
			s_wActiveLed = 0;
			break;
		}
	}
}

/* ------------------------------------------------------------------------- */


void SendString(const char *pStr,char bShowCLRF)
{
	char nChar;
	char bCrNl=0;
	while (*pStr) {
		nChar=*pStr++;
		// handling for some special chars
		switch (nChar) {
			case '\\':	// command sequence is following
				nChar=*pStr++;
				switch (nChar) {
					case 'x':	// hex code following
						break;
					case 'n':	// send newline
						nChar='\n';
						break;
					case 't':	// send tab
						nChar='\t';
						break;
					case 'v':
						ShowVersion();
						return;
				}
				break;
			case '\n':
			case '\r':
				if (bCrNl) {
					bCrNl=0;
					if (bShowCLRF) 	{// 
        				nChar='\n'; 
						break;		
					}
					else {		// CRLF is only terminating the line -> don't report to PC
						continue;
					}
				}
				else {
					bCrNl=1;
					continue;
				}
				break;
		}
        /* use last key and not current key status in order to avoid lost
           changes in key status. */
		SendChar(nChar);
	}
	if (bCrNl) {		// pending CRLF -> send now
		SendChar('\n');
	}
}

void CheckFSandMacros()
{	
#if ENABLE_FS
	uchar res;
	for (;;) {
		res=f_mount(0,&MMCCard);
		if (FR_OK != res) {
			ShowError(ERR_FSERROR,res);
		}
		else {
			if (FR_OK==f_open(&f,"macros.txt",FA_READ)) {
				f_close(&f);
				return;
			}
			else
				ShowError(ERR_NOMACROTXT);
		}
	}
#endif
}

/* ------------------------------------------------------------------------- */

int	main(void)
{
	uchar   key;
	char szKeyMatrix[3];

//	wdt_enable(WDTO_2S);

    PORTB = 0xff;   /* activate all pull-ups */
    DDRB = 0;       /* all pins input */
    PORTC = 0xff;   /* activate all pull-ups */
    DDRC = 0;       /* all pins input */
	DDRD =0;
	PORTD = (1<<3);	// activate pullup for D3

	odDebugInit();
	sei();

	memset(szKeyMatrix,0,sizeof(szKeyMatrix));

#if ACTIVATE_USB
	usbInit();
    hardwareInit();
	usbDeviceConnect();
#endif
	CheckFSandMacros();

	for(;;){	/* main event loop */
		
		if (PIND & (1<<3)) {	// sd_card removed
			f_mount(0,NULL);	// unmount
			ShowError(ERR_NOCARD);
			CheckFSandMacros();
		}
		wdt_reset();
		#if ACTIVATE_USB
			usbPoll();
		#endif
		key = KeyLEDCyclic();
        if(key) {
			if (key <= 8) {
				szKeyMatrix[0]='A'+key-1;
				s_wActiveLed &= 0xff00;
			}
			else {
				szKeyMatrix[1]='0'+key-8;
				s_wActiveLed &= 0x00ff;
			}
			s_wActiveLed ^= (1<< (key-1) );
		
		#if TEST_CODE && ACTIVATE_USB
		
			int i;
			for (i=1;i<=56;i++) {
				TestKeyKode(i);	
			}
			TestKeyKode(100);	
			for (i=135;i<=152;i++) {
				TestKeyKode(i);	
			}

		#elif TEST_STRINGS

			char szTemp[10];
			const char *test0=" !\"#$%&()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_` abcdefghijklmnopqrstuvwxyz{|}~ ";
			const char *test1="Hallo ";
			const char *test2="Ines!";
			const char *test3="Noch mal hallo";
			const char *test4="40120,";
			const char *test5="Diagnose";
			const char *testleer="--";
			const char *p;

			switch (key) {
				case 0: p=test0; break;
				case 1: p=test1; break;
				case 2: p=test2; break;
				case 3: p=test3; break;
				case 4: p=test4; break;
				case 5: p=test5; break;
				default:p=testleer; break;
			}
			sprintf(szTemp,"%i:",key);
			SendString(szTemp);
			SendString(p);			

		#else
			if (szKeyMatrix[0] && szKeyMatrix[1]) {
				if (FindString(szKeyMatrix))
					SendString(s_szTempBuffer);
				else
					SendString(szKeyMatrix);
				memset(szKeyMatrix,0,sizeof(szKeyMatrix));
				s_wActiveLed = 0;
			}
		#endif

		}

	}
	return 0;
}

/* ------------------------------------------------------------------------- */

